# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from __future__ import print_function

import math
import warnings

import numpy as np
from numpy.linalg import inv
try:
    from scipy.optimize import brentq, leastsq, minimize
except ImportError:
    warnings.warn('Warning: no scipy')

import ase.units as units

from matscipy.atomic_strain import atomic_strain
from matscipy.elasticity import (rotate_elastic_constants,
                                 rotate_cubic_elastic_constants,
                                 Voigt_6_to_full_3x3_stress)
from matscipy.surface import MillerDirection, MillerPlane

from matscipy.neighbours import neighbour_list

###

# Constants
PLANE_STRAIN = 'plane strain'
PLANE_STRESS = 'plane stress'

MPa_sqrt_m = 1e6*units.Pascal*np.sqrt(units.m)

###

class RectilinearAnisotropicCrack:
    """
    Near field solution for a crack in a rectilinear anisotropic elastic medium.
    See:
    G. C. Sih, P. C. Paris and G. R. Irwin, Int. J. Frac. Mech. 1, 189 (1965)
    """
    def __init__(self):
        self.a11 = None
        self.a22 = None
        self.a12 = None
        self.a16 = None
        self.a26 = None
        self.a66 = None


    def set_plane_stress(self, a11, a22, a12, a16, a26, a66):
        self.a11 = a11
        self.a22 = a22
        self.a12 = a12
        self.a16 = a16
        self.a26 = a26
        self.a66 = a66

        self._init_crack()


    def set_plane_strain(self, b11, b22, b33, b12, b13, b23, b16, b26, b36,
                         b66):
        self.a11 = b11 - (b13*b13)/b33
        self.a22 = b22 - (b23*b23)/b33
        self.a12 = b12 - (b13*b23)/b33
        self.a16 = b16 - (b13*b36)/b33
        self.a26 = b26 - (b23*b36)/b33
        self.a66 = b66

        self._init_crack()


    def _init_crack(self):
        """
        Initialize dependent parameters.
        """
        p = np.poly1d( [ self.a11, -2*self.a16, 2*self.a12+self.a66,
                         -2*self.a26, self.a22 ] )

        mu1, mu2, mu3, mu4 = p.r

        if mu1 != mu2.conjugate():
            raise RuntimeError('Roots not in pairs.')

        if mu3 != mu4.conjugate():
            raise RuntimeError('Roots not in pairs.')

        self.mu1 = mu1
        self.mu2 = mu3

        self.p1  = self.a11*(self.mu1**2) + self.a12 - self.a16*self.mu1
        self.p2  = self.a11*(self.mu2**2) + self.a12 - self.a16*self.mu2

        self.q1  = self.a12*self.mu1 + self.a22/self.mu1 - self.a26
        self.q2  = self.a12*self.mu2 + self.a22/self.mu2 - self.a26

        self.inv_mu1_mu2 = 1/(self.mu1 - self.mu2)
        self.mu1_p2 = self.mu1 * self.p2
        self.mu2_p1 = self.mu2 * self.p1
        self.mu1_q2 = self.mu1 * self.q2
        self.mu2_q1 = self.mu2 * self.q1


    def displacements(self, r, theta, k):
        """
        Displacement field in mode I fracture. Positions are passed in cyclinder
        coordinates.

        Parameters
        ----------
        r : array_like
            Distances from the crack tip.
        theta : array_like
            Angles with respect to the plane of the crack.
        k : float
            Stress intensity factor.

        Returns
        -------
        u : array
            Displacements parallel to the plane of the crack.
        v : array
            Displacements normal to the plane of the crack.
        """

        h1 = k * np.sqrt(2.0*r/math.pi)

        h2 = np.sqrt( np.cos(theta) + self.mu2*np.sin(theta) )
        h3 = np.sqrt( np.cos(theta) + self.mu1*np.sin(theta) )

        u = h1*( self.inv_mu1_mu2*( self.mu1_p2*h2 - self.mu2_p1*h3 ) ).real
        v = h1*( self.inv_mu1_mu2*( self.mu1_q2*h2 - self.mu2_q1*h3 ) ).real

        return u, v


    def deformation_gradient(self, r, theta, k):
        """
        Deformation gradient tensor in mode I fracture. Positions are passed in
        cyclinder coordinates.

        Parameters
        ----------
        r : array_like
            Distances from the crack tip.
        theta : array_like
            Angles with respect to the plane of the crack.
        k : float
            Stress intensity factor.

        Returns
        -------
        du_dx : array
            Derivatives of displacements parallel to the plane within the plane.
        du_dy : array
            Derivatives of displacements parallel to the plane perpendicular to
            the plane.
        dv_dx : array
            Derivatives of displacements normal to the plane of the crack within
            the plane.
        dv_dy : array
            Derivatives of displacements normal to the plane of the crack
            perpendicular to the plane.
        """

        f = k / np.sqrt(2*math.pi*r)

        h1 = (self.mu1*self.mu2)*self.inv_mu1_mu2
        h2 = np.sqrt( np.cos(theta) + self.mu2*np.sin(theta) )
        h3 = np.sqrt( np.cos(theta) + self.mu1*np.sin(theta) )

        du_dx = f*( self.inv_mu1_mu2*( self.mu1_p2/h2 - self.mu2_p1/h3 ) ).real
        du_dy = f*( h1*( self.p2/h2 - self.p1/h3 ) ).real

        dv_dx = f*( self.inv_mu1_mu2*( self.mu1_q2/h2 - self.mu2_q1/h3 ) ).real
        dv_dy = f*( h1*( self.q2/h2 - self.q1/h3 ) ).real

        # We need to add unity matrix to turn this into the deformation gradient
        # tensor.
        du_dx += np.ones_like(du_dx)
        dv_dy += np.ones_like(dv_dy)

        return np.transpose([[du_dx, du_dy], [dv_dx, dv_dy]])


    def stresses(self, r, theta, k):
        """
        Stress field in mode I fracture. Positions are passed in cylinder
        coordinates.

        Parameters
        ----------
        r : array_like
            Distances from the crack tip.
        theta : array_like
            Angles with respect to the plane of the crack.
        k : float
            Stress intensity factor.

        Returns
        -------
        sig_x : array
            Diagonal component of stress tensor parallel to the plane of the
            crack.
        sig_y : array
            Diagonal component of stress tensor normal to the plane of the
            crack.
        sig_xy : array
            Off-diagonal component of the stress tensor.
        """

        f = k / np.sqrt(2.0*math.pi*r)

        h1 = (self.mu1*self.mu2)*self.inv_mu1_mu2
        h2 = np.sqrt( np.cos(theta) + self.mu2*np.sin(theta) )
        h3 = np.sqrt( np.cos(theta) + self.mu1*np.sin(theta) )

        sig_x  = f*(h1*(self.mu2/h2 - self.mu1/h3)).real
        sig_y  = f*(self.inv_mu1_mu2*(self.mu1/h2 - self.mu2/h3)).real
        sig_xy = f*(h1*(1/h3 - 1/h2)).real

        return sig_x, sig_y, sig_xy


    def _f(self, theta, v):
        h2 = ( cos(theta) + self.mu2*sin(theta) )**0.5
        h3 = ( cos(theta) + self.mu1*sin(theta) )**0.5

        return v - ( self.mu1_p2 * h2 - self.mu2_p1 * h3 ).real/ \
            ( self.mu1_q2 * h2 - self.mu2_q1 * h3 ).real


    def rtheta(self, u, v, k):
        """
        Invert displacement field in mode I fracture, i.e. compute r and theta
        from displacements.
        """

        # u/v = (self.mu1_p2*h2 - self.mu2_p1*h3)/(self.mu1_q2*h2-self.mu2_q1*h3)
        theta = brentq(self._f, -pi, pi, args=(u/v))

        h1 = k * sqrt(2.0*r/math.pi)

        h2 = ( cos(theta) + self.mu2*sin(theta) )**0.5
        h3 = ( cos(theta) + self.mu1*sin(theta) )**0.5

        sqrt_2_r = ( self.inv_mu1_mu2 * ( self.mu1_p2 * h2 - self.mu2_p1 * h3 ) ).real/k
        r1 = sqrt_2_r**2/2
        sqrt_2_r = ( self.inv_mu1_mu2 * ( self.mu1_q2 * h2 - self.mu2_q1 * h3 ) ).real/k
        r2 = sqrt_2_r**2/2

        return ( (r1+r2)/2, theta )


    def k1g(self, surface_energy):
        """
        K1G, Griffith critical stress intensity in mode I fracture
        """

        return math.sqrt(-4*surface_energy / \
                         (self.a22*
                          ((self.mu1+self.mu2)/(self.mu1*self.mu2)).imag))


    def k1gsqG(self):
        return -2/(self.a22*((self.mu1+self.mu2)/(self.mu1*self.mu2)).imag)

###

def displacement_residuals(r0, crack, x, y, ref_x, ref_y, k, power=1):
    """
    Return actual displacement field minus ideal displacement field,
    divided by r**alpha.
    """
    x0, y0 = r0
    u1x = x - ref_x
    u1y = y - ref_y
    dx = ref_x - x0
    dy = ref_y - y0
    abs_dr = np.sqrt(dx*dx+dy*dy)
    theta = np.arctan2(dy, dx)
    u2x, u2y = crack.displacements_from_cylinder_coordinates(abs_dr, theta, k)
    if abs(power) < 1e-12:
        power_of_abs_dr = 1.0
    else:
        power_of_abs_dr = abs_dr**power
    return (u1x - u2x)/power_of_abs_dr, (u1y - u2y)/power_of_abs_dr


def displacement_residual(r0, crack, x, y, ref_x, ref_y, k, mask=None, power=1):
    dux, duy = displacement_residuals(r0, crack, x, y, ref_x, ref_y, k,
                                      power=power)
    if mask is None:
        return np.transpose([dux, duy]).flatten()
    else:
        return np.transpose([dux[mask], duy[mask]]).flatten()


def deformation_gradient_residuals(r0, crack, x, y, cur, ref_x, ref_y, ref, k,
                                   cutoff):
    """
    Return actual displacement field minus ideal displacement field.
    """
    x0, y0 = r0

    cur = cur.copy()
    cur.set_positions(np.transpose([x, y, cur.positions[:, 2]]))
    ref = ref.copy()
    ref.set_positions(np.transpose([ref_x, ref_y, ref.positions[:, 2]]))

    F1, res1 = atomic_strain(cur, ref, cutoff=cutoff)
    # F1 is a 3x3 tensor, we throw away all z-components
    [F1xx, F1yx, F1zx], [F1xy, F1yy, F1zy], [F1xz, F1yz, F1zz] = F1.T
    F1 = np.array([[F1xx, F1xy], [F1yx, F1yy]]).T

    F2 = crack.deformation_gradient(ref.positions[:, 0], ref.positions[:, 1],
                                    x0, y0, k)

    return F1-F2


def deformation_gradient_residual(r0, crack, x, y, cur, ref_x, ref_y, ref, k,
                                  cutoff, mask=None):
    dF = deformation_gradient_residuals(r0, crack, x, y, cur, ref_x, ref_y, ref,
                                        k, cutoff)
    if mask is None:
        return dF.flatten()
    else:
        #return (dF[mask]*dF[mask]).sum(axis=2).sum(axis=1)
        return dF[mask].flatten()

###

class CubicCrystalCrack:
    """
    Crack in a cubic crystal.
    """

    def __init__(self, crack_surface, crack_front, C11=None, C12=None,
                 C44=None, stress_state=PLANE_STRAIN, C=None, Crot=None):
        """
        Initialize a crack in a cubic crystal with elastic constants C11, C12
        and C44 (or optionally a full 6x6 elastic constant matrix C).
        The crack surface is given by crack_surface, the cracks runs
        in the plane given by crack_front.
        """

        # x (third_dir) - direction in which the crack is running
        # y (crack_surface) - free surface that forms due to the crack
        # z (crack_front) - direction of the crack front
        third_dir = np.cross(crack_surface, crack_front)
        third_dir = np.array(third_dir) / np.sqrt(np.dot(third_dir,
                                                         third_dir))
        crack_surface = np.array(crack_surface) / \
            np.sqrt(np.dot(crack_surface, crack_surface))
        crack_front = np.array(crack_front) / \
            np.sqrt(np.dot(crack_front, crack_front))

        A = np.array([third_dir, crack_surface, crack_front])
        if np.linalg.det(A) < 0:
            third_dir = -third_dir
        A = np.array([third_dir, crack_surface, crack_front])

        if Crot is not None:
            C6 = Crot
        elif C is not None:
            C6 = rotate_elastic_constants(C, A)
        else:
            C6 = rotate_cubic_elastic_constants(C11, C12, C44, A)

        self.crack = RectilinearAnisotropicCrack()
        S6 = inv(C6)

        self.C = C6
        self.S = S6

        if stress_state == PLANE_STRESS:
            self.crack.set_plane_stress(S6[0, 0], S6[1, 1], S6[0, 1],
                                        S6[0, 5], S6[1, 5], S6[5, 5])
        elif stress_state == PLANE_STRAIN:
            self.crack.set_plane_strain(S6[0, 0], S6[1, 1], S6[2, 2],
                                        S6[0, 1], S6[0, 2], S6[1, 2],
                                        S6[0, 5], S6[1, 5], S6[2, 5],
                                        S6[5, 5])


    def k1g(self, surface_energy):
        """
        Compute Griffith critical stress intensity in mode I fracture.

        Parameters
        ----------
        surface_energy : float
            Surface energy of the respective crystal surface.

        Returns
        -------
        k1g : float
            Stress intensity factor.
        """
        return self.crack.k1g(surface_energy)


    def k1gsqG(self):
        return self.crack.k1gsqG()


    def displacements_from_cylinder_coordinates(self, r, theta, k):
        """
        Displacement field in mode I fracture from cylindrical coordinates.
        """
        return self.crack.displacements(r, theta, k)


    def displacements_from_cartesian_coordinates(self, dx, dy, k):
        """
        Displacement field in mode I fracture from cartesian coordinates.
        """
        abs_dr = np.sqrt(dx*dx+dy*dy)
        theta = np.arctan2(dy, dx)
        return self.displacements_from_cylinder_coordinates(abs_dr, theta, k)


    def displacements(self, ref_x, ref_y, x0, y0, k):
        """
        Displacement field for a list of cartesian positions.

        Parameters
        ----------
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        x0 : float
            x-coordinate of the crack tip.
        y0 : float
            y-coordinate of the crack tip.
        k : float
            Stress intensity factor.

        Returns
        -------
        ux : array_like
            x-displacements.
        uy : array_like
            y-displacements.
        """
        dx = ref_x - x0
        dy = ref_y - y0
        return self.displacements_from_cartesian_coordinates(dx, dy, k)


    def deformation_gradient_from_cylinder_coordinates(self, r, theta, k):
        """
        Displacement field in mode I fracture from cylindrical coordinates.
        """
        return self.crack.deformation_gradient(r, theta, k)


    def deformation_gradient_from_cartesian_coordinates(self, dx, dy, k):
        """
        Displacement field in mode I fracture from cartesian coordinates.
        """
        abs_dr = np.sqrt(dx*dx+dy*dy)
        theta = np.arctan2(dy, dx)
        return self.deformation_gradient_from_cylinder_coordinates(abs_dr, theta, k)


    def deformation_gradient(self, ref_x, ref_y, x0, y0, k):
        """
        Deformation gradient for a list of cartesian positions.

        Parameters
        ----------
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        x0 : float
            x-coordinate of the crack tip.
        y0 : float
            y-coordinate of the crack tip.
        k : float
            Stress intensity factor.

        Returns
        -------
        ux : array_like
            x-displacements.
        uy : array_like
            y-displacements.
        """
        dx = ref_x - x0
        dy = ref_y - y0
        return self.deformation_gradient_from_cartesian_coordinates(dx, dy, k)


    def crack_tip_position(self, x, y, ref_x, ref_y, x0, y0, k, mask=None,
                           residual_func=displacement_residual,
                           method='Powell', return_residuals=False):
        """
        Return an estimate of the real crack tip position by minimizing the
        mean square error of the current configuration relative to the
        diplacement field obtained for a stress intensity factor k from linear
        elastic fracture mechanics.

        Parameters
        ----------
        x : array_like
            x-positions of the atomic system containing the crack.
        y : array_like
            y-positions of the atomic system containing the crack.
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        x0 : float
            Initial guess for the x-coordinate of the crack tip.
        y0 : float
            Initial guess for the y-coordinate of the crack tip.
        k : float
            Stress intensity factor.
        mask : array_like, optional
            Marks the atoms to use for this calculation.
        residual_func : function
            Function returning per-atom residuals to be minimized.
        method : str
            Optimization method. See method argument of scipy.optimize.minimize.
            Additionally, 'leastsq' invokes scipy.optimize.leastsq.
        return_residuals : bool
            Function returns residuals if set to True.

        Returns
        -------
        x0 : float
            x-coordinate of the crack tip.
        y0 : float
            y-coordinate of the crack tip.
        residuals : array, optional
            Per-atom residuals at end of optimization.
        """
        if mask is None:
            mask = np.ones(len(x), dtype=bool)
        if method == 'leastsq':
            (x1, y1), ier = leastsq(residual_func, (x0, y0),
                                    args=(self, x, y, ref_x, ref_y, k, mask))
            if ier not in [ 1, 2, 3, 4 ]:
                raise RuntimeError('Could not find crack tip')
        else:
            opt = minimize(lambda *args: (residual_func(*args)**2).sum(),
                           (x0, y0), args=(self, x, y, ref_x, ref_y, k, mask),
                           method=method)
            if not opt.success:
                raise RuntimeError('Could not find crack tip. Reason: {}'
                                   .format(opt.message))
            x1, y1 = opt.x
        if return_residuals:
            return x1, y1, residual_func((x0, y0), self, x, y, ref_x, ref_y, k)
        else:
            return x1, y1


    def _residual_y(self, y0, x0, x, y, ref_x, ref_y, k, mask):
        dux, duy = self.displacement_residuals(x, y, ref_x, ref_y, x0, y0, k)
        return dux[mask]*dux[mask]+duy[mask]*duy[mask]


    def crack_tip_position_y(self, x, y, ref_x, ref_y, x0, y0, k, mask=None):
        """
        Return an estimate of the y-coordinate of the real crack tip position
        assuming the stress intensity factor is k.

        Parameters
        ----------
        x : array_like
            x-positions of the atomic system containing the crack.
        y : array_like
            y-positions of the atomic system containing the crack.
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        x0 : float
            Initial guess for the x-coordinate of the crack tip.
        y0 : float
            Initial guess for the y-coordinate of the crack tip.
        k : float
            Stress intensity factor.
        mask : array_like, optional
            Marks the atoms to use for this calculation.

        Returns
        -------
        y0 : float
            y-coordinate of the crack tip
        """
        if mask is None:
            mask = np.ones(len(a), dtype=bool)
        ( y0, ), ier = leastsq(self._residual_y, y0,
                          args=(x0, x, y, ref_x, ref_y, k, mask))
        if ier not in [ 1, 2, 3, 4 ]:
            raise RuntimeError('Could not find crack tip')
        return y0


    def scale_displacements(self, x, y, ref_x, ref_y, old_k, new_k):
        """
        Rescale atomic positions from stress intensity factor old_k to the new
        stress intensity factor new_k. This is useful for extrapolation of
        relaxed positions.

        Parameters
        ----------
        x : array_like
            x-positions of the atomic system containing the crack.
        y : array_like
            y-positions of the atomic system containing the crack.
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        old_k : float
            Current stress intensity factor.
        new_k : float
            Stress intensity factor for the output positions.

        Returns
        -------
        x : array_like
            New x-positions of the atomic system containing the crack.
        y : array_like
            New y-positions of the atomic system containing the crack.
        """
        return ref_x + new_k/old_k*(x-ref_x), ref_y + new_k/old_k*(y-ref_y)


    def stresses_from_cylinder_coordinates(self, r, theta, k):
        """
        Stress field in mode I fracture from cylindrical coordinates
        """
        return self.crack.stresses(r, theta, k)


    def stresses_from_cartesian_coordinates(self, dx, dy, k):
        """
        Stress field in mode I fracture from cartesian coordinates
        """
        abs_dr = np.sqrt(dx*dx+dy*dy)
        theta = np.arctan2(dy, dx)
        return self.stresses_from_cylinder_coordinates(abs_dr, theta, k)


    def stresses(self, ref_x, ref_y, x0, y0, k):
        """
        Stress field for a list of cartesian positions

        Parameters
        ----------
        ref_x : array_like
            x-positions of the reference crystal.
        ref_y : array_like
            y-positions of the reference crystal.
        x0 : float
            x-coordinate of the crack tip.
        y0 : float
            y-coordinate of the crack tip.
        k : float
            Stress intensity factor.

        Returns
        -------
        sig_x : array_like
            xx-component of the stress tensor

        sig_y : array_like
            yy-component of the stress tensor

        sig_xy : array_like
            xy-component of the stress tensor
        """
        dx = ref_x - x0
        dy = ref_y - y0
        sig_x, sig_y, sig_xy = self.stresses_from_cartesian_coordinates(dx, dy, k)
        return sig_x, sig_y, sig_xy



def isotropic_modeI_crack_tip_stress_field(K, r, t, xy_only=True,
                                           nu=0.5, stress_state=PLANE_STRAIN):
    """
    Compute Irwin singular crack tip stress field for mode I fracture.

    Parameters
    ----------
    K : float
       Mode I stress intensity factor. Units should match units of `r`.
    r : array_like
       Radial distances from crack tip. Can be a multidimensional
       array to evaluate stress field on a grid.
    t : array_like
       Angles from horzontal line y=0 ahead of crack tip,
       measured anticlockwise. Should have same shape as `r`.
    xy_only : bool
       If True (default) only xx, yy, xy and yx components will be set.
    nu : float
       Poisson ratio. Used only when ``xy_only=False``, to determine zz stresses
    stress_state : str
       One of"plane stress" or "plane strain". Used if xyz_only=False to
       determine zz stresses.

    Returns
    -------
    sigma : array with shape ``r.shape + (3,3)``
    """

    if r.shape != t.shape:
        raise ValueError('Shapes of radial and angular arrays "r" and "t" '
                         'must match.')

    if stress_state not in [PLANE_STRAIN, PLANE_STRESS]:
        raise ValueError('"stress_state" should be either "{0}" or "{1}".'
            .format(PLANE_STRAIN, PLANE_STRESS))

    sigma = np.zeros(r.shape + (3, 3))
    radial = K/np.sqrt(2*math.pi*r)

    sigma[...,0,0] = radial*np.cos(t/2.0)*(1.0 - np.sin(t/2.0)*np.sin(3.0*t/2.0)) # xx
    sigma[...,1,1] = radial*np.cos(t/2.0)*(1.0 + np.sin(t/2.0)*np.sin(3.0*t/2.0)) # yy
    sigma[...,0,1] = radial*np.sin(t/2.0)*np.cos(t/2.0)*np.cos(3.0*t/2.0)         # xy
    sigma[...,1,0] = sigma[...,0,1]                                               # yx=xy

    if not xy_only and stress_state == PLANE_STRAIN:
        sigma[...,2,2] = nu*(sigma[...,0,0] + sigma[...,1,1])              # zz

    return sigma


def isotropic_modeI_crack_tip_displacement_field(K, G, nu, r, t,
                                                 stress_state=PLANE_STRAIN):
    """
    Compute Irwin singular crack tip displacement field for mode I fracture.

    Parameters
    ----------
    K : float
        Mode I stress intensity factor. Units should match units of `G` and `r`.
    G : float
        Shear modulus. Units should match units of `K` and `r`.
    nu : float
        Poisson ratio.
    r : array_like
        Radial distances from crack tip. Can be a multidimensional
        array to evaluate stress field on a grid.
    t : array_like
        Angles from horizontal line y=0 ahead of crack tip,
        measured anticlockwise. Should have same shape as `r`.
    stress_state : str
        One of"plane stress" or "plane strain". Used if xyz_only=False to
        determine zz stresses.

    Returns
    -------
    u : array
    v : array
        Displacements. Same shape as `r` and `t`.
    """

    if r.shape != t.shape:
        raise ValueError('Shapes of radial and angular arrays "r" and "t" '
                         'must match.')

    if stress_state == PLANE_STRAIN:
        kappa = 3-4*nu
    elif stress_state == PLANE_STRESS:
        kappa = (3.-nu)/(1.+nu)
    else:
        raise ValueError('"stress_state" should be either "{0}" or "{1}".'
            .format(PLANE_STRAIN, PLANE_STRESS))

    radial = K*np.sqrt(r/(2.*math.pi))/(2.*G)
    u = radial*np.cos(t/2)*(kappa-1+2*np.sin(t/2)**2)
    v = radial*np.sin(t/2)*(kappa+1-2*np.cos(t/2)**2)

    # Form in Lawn book is equivalent:
    #radial = K/(4*G)*np.sqrt(r/(2.*math.pi))
    #u = radial*((2*kappa - 1)*np.cos(t/2) - np.cos(3*t/2))
    #v = radial*((2*kappa + 1)*np.sin(t/2) - np.sin(3*t/2))

    return u, v


class IsotropicStressField(object):
    """
    Calculator to return Irwin near-tip stress field at atomic sites
    """
    def __init__(self, K=None, x0=None, y0=None, sxx0=0.0, syy0=0., sxy0=0., nu=0.5,
                 stress_state='plane strain'):
        self.K = K
        self.x0 = x0
        self.y0 = y0
        self.sxx0 = sxx0
        self.syy0 = syy0
        self.sxy0 = sxy0
        self.nu = nu
        self.stress_state = stress_state

    def get_stresses(self, atoms):
        K = self.K
        if K is None:
            K = get_stress_intensity_factor(atoms)

        x0, y0 = self.x0, self.y0
        if x0 is None:
            x0 = atoms.info['CrackPos'][0]
        if y0 is None:
            y0 = atoms.info['CrackPos'][1]

        x = atoms.positions[:, 0]
        y = atoms.positions[:, 1]
        r = np.sqrt((x - x0)**2 + (y - y0)**2)
        t = np.arctan2(y - y0, x - x0)

        sigma = isotropic_modeI_crack_tip_stress_field(K, r, t, self.nu,
                                                       self.stress_state)
        sigma[:,0,0] += self.sxx0
        sigma[:,1,1] += self.syy0
        sigma[:,0,1] += self.sxy0
        sigma[:,1,0] += self.sxy0

        return sigma


def strain_to_G(strain, E, nu, orig_height):
    """
    Convert from strain to energy release rate G for thin strip geometry

    Parameters
    ----------
    strain : float
       Dimensionless ratio ``(current_height - orig_height)/orig_height``
    E : float
       Young's modulus relevant for a pull in y direction sigma_yy/eps_yy
    nu : float
       Poission ratio -eps_yy/eps_xx
    orig_height : float
       Unstrained height of slab

    Returns
    -------
    G : float
       Energy release rate in units consistent with input
       (i.e. in eV/A**2 if eV/A/fs units used)
    """
    return 0.5 * E / (1.0 - nu * nu) * strain * strain * orig_height


def G_to_strain(G, E, nu, orig_height):
    """
    Convert from energy release rate G to strain for thin strip geometry

    Parameters
    ----------
    G : float
       Energy release rate in units consistent with `E` and `orig_height`
    E : float
       Young's modulus relevant for a pull in y direction sigma_yy/eps_yy
    nu : float
       Poission ratio -eps_yy/eps_xx
    orig_height : float
       Unstrained height of slab

    Returns
    -------
    strain : float
       Dimensionless ratio ``(current_height - orig_height)/orig_height``
    """
    return np.sqrt(2.0 * G * (1.0 - nu * nu) / (E * orig_height))


def get_strain(atoms):
    """
    Return the current strain on thin strip configuration `atoms`

    Requires unstrained height of slab to be stored as ``OrigHeight``
    key in ``atoms.info`` dictionary.

    Also updates value stored in ``atoms.info``.
    """

    orig_height = atoms.info['OrigHeight']
    current_height = atoms.positions[:, 1].max() - atoms.positions[:, 1].min()
    strain = current_height / orig_height - 1.0
    atoms.info['strain'] = strain
    return strain


def get_energy_release_rate(atoms):
    """
    Return the current energy release rate G for `atoms`

    Result is computed assuming thin strip geometry, and using
    stored Young's modulus and Poission ratio and original slab height
    from `atoms.info` dictionary.

    Also updates `G` value stored in ``atoms.info`` dictionary.
    """

    current_strain = get_strain(atoms)
    orig_height = atoms.info['OrigHeight']
    E = atoms.info['YoungsModulus']
    nu = atoms.info['PoissonRatio_yx']
    G = strain_to_G(current_strain, E, nu, orig_height)
    atoms.info['G'] = G
    return G


def get_stress_intensity_factor(atoms, stress_state=PLANE_STRAIN):
    """
    Compute stress intensity factor K_I

    Calls :func:`get_energy_release_rate` to compute `G`, then
    uses stored `YoungsModulus` and `PoissionRatio_yz` values from
    `atoms.info` dictionary to compute K_I.

    Also updates value stored in ``atoms.info`` dictionary.
    """

    G = get_energy_release_rate(atoms)

    E = atoms.info['YoungsModulus']
    nu = atoms.info['PoissonRatio_yx']

    if stress_state == PLANE_STRAIN:
        Ep = E/(1-nu**2)
    elif stress_state == PLANE_STRESS:
        Ep = E
    else:
        raise ValueError('"stress_state" should be either "{0}" or "{1}".'
            .format(PLANE_STRAIN, PLANE_STRESS))

    K = np.sqrt(G*Ep)
    atoms.info['K'] = K
    return K


def fit_crack_stress_field(atoms, r_range=(0., 50.), initial_params=None, fix_params=None,
                           sigma=None, avg_sigma=None, avg_decay=0.005, calc=None, verbose=False):
    """
    Perform a least squares fit of near-tip stress field to isotropic solution

    Stresses on the atoms are fit to the Irwin K-field singular crack tip
    solution, allowing the crack position, stress intensity factor and
    far-field stress components to vary during the fit.

    Parameters
    ----------
    atoms : :class:`~.Atoms` object
       Crack system. For the initial fit, the following keys are used
       from the :attr:`~Atoms.info` dictionary:

          - ``YoungsModulus``
          - ``PossionRatio_yx``
          - ``G`` --- current energy release rate
          - ``strain`` --- current applied strain
          - ``CrackPos`` --- initial guess for crack tip position

       The initial guesses for the stress intensity factor ``K`` and
       far-field stress ``sigma0`` are computed from
       ``YoungsModulus``, ``PoissonRatio_yx``, ``G`` and ``strain``,
       assuming plane strain in thin strip boundary conditions.

       On exit, new ``K``, ``sigma0`` and ``CrackPos`` entries are set
       in the :attr:`~Atoms.info` dictionary. These values are then
       used as starting guesses for subsequent fits.

    r_range : sequence of two floats, optional
       If present, restrict the stress fit to an annular region
       ``r_range[0] <= r < r_range[1]``, centred on the previous crack
       position (from the ``CrackPos`` entry in ``atoms.info``). If
       r_range is ``None``, fit is carried out for all atoms.

    initial_params : dict
       Names and initial values of parameters. Missing initial values
       are guessed from Atoms object.

    fix_params : dict
       Names and values of parameters to fix during the fit,
       e.g. ``{y0: 0.0}`` to constrain the fit to the line y=0

    sigma : None or array with shape (len(atoms), 3, 3)
       Explicitly provide the per-atom stresses. Avoids calling Atoms'
       calculators :meth:`~.get_stresses` method.

    avg_sigma : None or array with shape (len(atoms), 3, 3)
       If present, use this array to accumulate the time-averaged
       stress field. Useful when processing a trajectory.

    avg_decay : real
       Factor by which average stress is attenuated at each step.
       Should be set to ``dt/tau`` where ``dt`` is MD time-step
       and ``tau`` is a characteristic averaging time.

    calc : Calculator object, optional
       If present, override the calculator used to compute stresses
       on the atoms. Default is ``atoms.get_calculator``.

       To use the atom resolved stress tensor pass an instance of the
       :class:`~quippy.elasticity.AtomResolvedStressField` class.

    verbose : bool, optional
       If set to True, print additional information about the fit.

    Returns
    -------
    params : dict with keys ``[K, x0, y0, sxx0, syy0, sxy0]``
       Fitted parameters, in a form suitable for passing to
       :class:`IsotropicStressField` constructor. These are the stress intensity
       factor `K`, the centre of the stress field ``(x0, y0)``, and the
       far field contribution to the stress ``(sxx0, syy0, sxy0)``.
    """

    params = {}
    if initial_params is not None:
       params.update(initial_params)

    if 'K' not in params:
       # Guess for stress intensity factor K
       if 'K' in atoms.info:
           params['K'] = atoms.info['K']
       else:
           try:
               params['K'] = get_stress_intensity_factor(atoms)
           except KeyError:
               params['K'] = 1.0*MPa_sqrt_m

    if 'sxx0' not in params or 'syy0' not in params or 'sxy0' not in params:
       # Guess for far-field stress
       if 'sigma0' in atoms.info:
          params['sxx0'], params['syy0'], params['sxy0'] = atoms.info['sigma0']
       else:
          try:
              E = atoms.info['YoungsModulus']
              nu = atoms.info['PoissonRatio_yx']
              Ep = E/(1-nu**2)
              params['syy0'] = Ep*atoms.info['strain']
              params['sxx0'] = nu*params['syy0']
              params['sxy0'] = 0.0
          except KeyError:
              params['syy0'] = 0.0
              params['sxx0'] = 0.0
              params['sxy0'] = 0.0

    if 'x0' not in params or 'y0' not in params:
       # Guess for crack position
       try:
           params['x0'], params['y0'], _ = atoms.info['CrackPos']
       except KeyError:
           params['x0'] = (atoms.positions[:, 0].min() +
                           (atoms.positions[:, 0].max() - atoms.positions[:, 0].min())/3.0)
           params['y0'] = 0.0

    # Override any fixed parameters
    if fix_params is None:
       fix_params = {}
    params.update(fix_params)

    x = atoms.positions[:, 0]
    y = atoms.positions[:, 1]
    r = np.sqrt((x - params['x0'])**2 + (y - params['y0'])**2)

    # Get local stresses
    if sigma is None:
       if calc is None:
           calc = atoms.get_calculator()
       sigma = calc.get_stresses(atoms)

    if sigma.shape != (len(atoms), 3, 3):
        sigma = Voigt_6_to_full_3x3_stress(sigma)

    # Update avg_sigma in place
    if avg_sigma is not None:
       avg_sigma[...] = np.exp(-avg_decay)*avg_sigma + (1.0 - np.exp(-avg_decay))*sigma
       sigma = avg_sigma.copy()

    # Zero components out of the xy plane
    sigma[:,2,2] = 0.0
    sigma[:,0,2] = 0.0
    sigma[:,2,0] = 0.0
    sigma[:,1,2] = 0.0
    sigma[:,2,1] = 0.0

    mask = Ellipsis # all atoms
    if r_range is not None:
        rmin, rmax = r_range
        mask = (r > rmin) & (r < rmax)

    if verbose:
       print('Fitting on %r atoms' % sigma[mask,1,1].shape)

    def objective_function(params, x, y, sigma, var_params):
        params = dict(zip(var_params, params))
        if fix_params is not None:
            params.update(fix_params)
        isotropic_sigma = IsotropicStressField(**params).get_stresses(atoms)
        delta_sigma = sigma[mask,:,:] - isotropic_sigma[mask,:,:]
        return delta_sigma.reshape(delta_sigma.size)

    # names and values of parameters which can vary in this fit
    var_params = sorted([key for key in params.keys() if key not in fix_params.keys() ])
    initial_params = [params[key] for key in var_params]

    from scipy.optimize import leastsq
    fitted_params, cov, infodict, mesg, success = leastsq(objective_function,
                                                         initial_params,
                                                         args=(x, y, sigma, var_params),
                                                         full_output=True)

    params = dict(zip(var_params, fitted_params))
    params.update(fix_params)

    # estimate variance in parameter estimates
    if cov is None:
       # singular covariance matrix
       err = dict(zip(var_params, [0.]*len(fitted_params)))
    else:
       s_sq = (objective_function(fitted_params, x, y, sigma, var_params)**2).sum()/(sigma.size-len(fitted_params))
       cov = cov * s_sq
       err = dict(zip(var_params, np.sqrt(np.diag(cov))))

    if verbose:
       print('K = %.3f MPa sqrt(m)' % (params['K']/MPA_SQRT_M))
       print('sigma^0_{xx,yy,xy} = (%.1f, %.1f, %.1f) GPa' % (params['sxx0']*GPA,
                                                              params['syy0']*GPA,
                                                              params['sxy0']*GPA))
       print('Crack position (x0, y0) = (%.1f, %.1f) A' % (params['x0'], params['y0']))

    atoms.info['K'] = params['K']
    atoms.info['sigma0'] = (params['sxx0'], params['syy0'], params['sxy0'])
    atoms.info['CrackPos'] = np.array((params['x0'], params['y0'], atoms.cell[2,2]/2.0))

    return params, err


def find_tip_coordination(a, bondlength=2.6, bulk_nn=4):
    """
    Find position of tip in crack cluster from coordination
    """
    i, j = neighbour_list("ij", a, bondlength)
    nn = np.bincount(i, minlength=len(a))

    a.set_array('n_neighb', nn)
    g = a.get_array('groups')

    y = a.positions[:, 1]
    above = (nn < bulk_nn) & (g != 0) & (y > a.cell[1,1]/2.0)
    below = (nn < bulk_nn) & (g != 0) & (y < a.cell[1,1]/2.0)

    a.set_array('above', above)
    a.set_array('below', below)

    bond1 = np.asscalar(above.nonzero()[0][a.positions[above, 0].argmax()])
    bond2 = np.asscalar(below.nonzero()[0][a.positions[below, 0].argmax()])

    # These need to be ints, otherwise they are no JSON serializable.
    a.info['bond1'] = bond1
    a.info['bond2'] = bond2

    return bond1, bond2


def find_tip_broken_bonds(atoms, cutoff, bulk_nn=4, boundary_thickness=None):
    """
    Find position of the tip from the atom coordination, i.e. broken bonds.
    Using the C implementation of 'neighbour_list'.
    Returns the tip's position in cartesian coordinates.

    Parameters
    ----------
    atoms : ase.Atoms
        Atomic configuration.
    cutoff : float
        Cutoff distance for neighbour search.
    bulk_nn : integer
        Number of nearest neighbours for the standard bulk configuration.
    boundary_buffer : float
        Thickness of the boundaries.
        Defaults to cutoff distance.

    Returns
    -------
    tip_position : numpy array
        The x and y values are found.
        The z value is calculated as the midpoint of the depth.
    """

    # initialisation of the boundaries
    if boundary_thickness is None:
        boundary_thickness = cutoff

    right_boundary = atoms.positions[(np.argmax(atoms.positions[:,0], axis=0)), 0] - boundary_thickness
    top_boundary = atoms.positions[(np.argmax(atoms.positions[:,1], axis=0)), 1] - boundary_thickness
    bottom_boundary = atoms.positions[(np.argmin(atoms.positions[:,1], axis=0)), 1] + boundary_thickness
    left_boundary = atoms.positions[(np.argmin(atoms.positions[:,0], axis=0)), 0] + boundary_thickness

    # calculating the coordination from the neighbours list
    i = neighbour_list("i", atoms, cutoff)
    coordination_list = np.bincount(i, minlength=len(atoms))

    # list of atom numbers with at least one broken bond
    broken_bonds_array = np.where(coordination_list <= bulk_nn-1)

    # finds the atom number with the most positive x-valued position with a broken bond(s)
    # within the bounded section
    atom_number = 0
    for m in range(0, len(broken_bonds_array[0])):
        temp_atom_pos = atoms.positions[broken_bonds_array[0][m]]
        if temp_atom_pos[0] > atoms.positions[atom_number,0]:
            if left_boundary < temp_atom_pos[0] < right_boundary:
                if bottom_boundary < temp_atom_pos[1] < top_boundary:
                    atom_number = m

    tip_position = atoms.positions[broken_bonds_array[0][atom_number]]

    return np.array((tip_position[0], tip_position[1], atoms.cell[2,2]/2.0))


def find_tip_stress_field(atoms, r_range=None, initial_params=None, fix_params=None,
                                sigma=None, avg_sigma=None, avg_decay=0.005, calc=None):
    """
    Find the position of crack tip by fitting to the isotropic `K`-field stress

    Fit is carried out using :func:`fit_crack_stress_field`, and parameters
    have the same meaning as there.

    See also
    --------
    fit_crack_stress_field
    """

    params, err = fit_crack_stress_field(atoms, r_range, initial_params, fix_params, sigma,
                                         avg_sigma, avg_decay, calc)

    return np.array((params['x0'], params['y0'], atoms.cell[2,2]/2.0))


def plot_stress_fields(atoms, r_range=None, initial_params=None, fix_params=None,
                       sigma=None, avg_sigma=None, avg_decay=0.005, calc=None):
    """
    Fit and plot atomistic and continuum stress fields

    Firstly a fit to the Irwin `K`-field solution is carried out using
    :func:`fit_crack_stress_field`, and parameters have the same
    meaning as for that function. Then plots of the
    :math:`\sigma_{xx}`, :math:`\sigma_{yy}`, :math:`\sigma_{xy}`
    fields are produced for atomistic and continuum cases, and for the
    residual error after fitting.
    """

    from pylab import griddata, meshgrid, subplot, cla, contourf, colorbar, draw, title, clf, gca

    params, err = fit_crack_stress_field(atoms, r_range, initial_params, fix_params, sigma,
                                         avg_sigma, avg_decay, calc)

    K, x0, y0, sxx0, syy0, sxy0 = (params['K'], params['x0'], params['y0'],
                                   params['sxx0'], params['syy0'], params['sxy0'])

    x = atoms.positions[:, 0]
    y = atoms.positions[:, 1]

    X = np.linspace((x-x0).min(), (x-x0).max(), 500)
    Y = np.linspace((y-y0).min(), (y-y0).max(), 500)

    t = np.arctan2(y-y0, x-x0)
    r = np.sqrt((x-x0)**2 + (y-y0)**2)

    if r_range is not None:
       rmin, rmax = r_range
       mask = (r > rmin) & (r < rmax)
    else:
       mask = Ellipsis

    atom_sigma = sigma
    if atom_sigma is None:
       atom_sigma = atoms.get_stresses()

    grid_sigma = np.dstack([griddata(x[mask]-x0, y[mask]-y0, atom_sigma[mask,0,0], X, Y),
                            griddata(x[mask]-x0, y[mask]-y0, atom_sigma[mask,1,1], X, Y),
                            griddata(x[mask]-x0, y[mask]-y0, atom_sigma[mask,0,1], X, Y)])

    X, Y = meshgrid(X, Y)
    R = np.sqrt(X**2+Y**2)
    T = np.arctan2(Y, X)

    grid_sigma[((R < rmin) | (R > rmax)),:] = np.nan # mask outside fitting region

    isotropic_sigma = isotropic_modeI_crack_tip_stress_field(K, R, T, x0, y0)
    isotropic_sigma[...,0,0] += sxx0
    isotropic_sigma[...,1,1] += syy0
    isotropic_sigma[...,0,1] += sxy0
    isotropic_sigma[...,1,0] += sxy0
    isotropic_sigma = ma.masked_array(isotropic_sigma, mask=grid_sigma.mask)

    isotropic_sigma[((R < rmin) | (R > rmax)),:,:] = np.nan # mask outside fitting region

    contours = [np.linspace(0, 20, 10),
                np.linspace(0, 20, 10),
                np.linspace(-10,10, 10)]

    dcontours = [np.linspace(0, 5, 10),
                np.linspace(0, 5, 10),
                np.linspace(-5, 5, 10)]

    clf()
    for i, (ii, jj), label in zip(range(3),
                                  [(0,0), (1,1), (0,1)],
                                  ['\sigma_{xx}', r'\sigma_{yy}', r'\sigma_{xy}']):
        subplot(3,3,i+1)
        gca().set_aspect('equal')
        contourf(X, Y, grid_sigma[...,i]*GPA, contours[i])
        colorbar()
        title(r'$%s^\mathrm{atom}$' % label)
        draw()

        subplot(3,3,i+4)
        gca().set_aspect('equal')
        contourf(X, Y, isotropic_sigma[...,ii,jj]*GPA, contours[i])
        colorbar()
        title(r'$%s^\mathrm{Isotropic}$' % label)
        draw()

        subplot(3,3,i+7)
        gca().set_aspect('equal')
        contourf(X, Y, abs(grid_sigma[...,i] -
                           isotropic_sigma[...,ii,jj])*GPA, dcontours[i])
        colorbar()
        title(r'$|%s^\mathrm{atom} - %s^\mathrm{isotropic}|$' % (label, label))
        draw()



def thin_strip_displacement_y(x, y, strain, a, b):
    """
    Return vertical displacement ramp used to apply initial strain to slab

    Strain is increased from 0 to strain over distance :math:`a <= x <= b`.
    Region :math:`x < a` is rigidly shifted up/down by ``strain*height/2``.

    Here is an example of how to use this function on an artificial
    2D square atomic lattice. The positions are plotted before (left)
    and after (right) applying the displacement, and the horizontal and
    vertical lines show the `strain` (red), `a` (green) and `b` (blue)
    parameters. ::

       import matplotlib.pyplot as plt
       import numpy as np

       w = 1; h = 1; strain = 0.1; a = -0.5; b =  0.0
       x = np.linspace(-w, w, 20)
       y = np.linspace(-h, h, 20)
       X, Y = np.meshgrid(x, y)
       u_y = thin_strip_displacement_y(X, Y, strain, a, b)

       for i, disp in enumerate([0, u_y]):
           plt.subplot(1,2,i+1)
           plt.scatter(X, Y + disp, c='k', s=5)
           for y in [-h, h]:
               plt.axhline(y, color='r', linewidth=2, linestyle='dashed')
               plt.axhline(y*(1+strain), color='r', linewidth=2)
           for x, c in zip([a, b], ['g', 'b']):
               plt.axvline(x, color=c, linewidth=2)

    .. image:: thin-strip-displacement-y.png
       :width: 600
       :align: center

    Parameters
    ----------
    x : array
    y : array
       Atomic positions in unstrained slab, centered on origin x=0,y=0
    strain : float
       Far field strain to apply
    a : float
       x coordinate for beginning of strain ramp
    b : float
       x coordinate for end of strain ramp
    """

    u_y = np.zeros_like(y)
    height = y.max() - y.min()     # measure height of slab
    shift = strain * height / 2.0  # far behind crack, shift = strain*height/2

    u_y[x < a] = np.sign(y[x < a]) * shift  # region shift for x < a
    u_y[x > b] = strain * y[x > b]          # constant strain for x > b

    middle = (x >= a) & (x <= b)            # interpolate for a <= x <= b
    f = (x[middle] - a) / (b - a)
    u_y[middle] = (f * strain * y[middle] +
                   (1 - f) * shift * np.sign(y[middle]))

    return u_y


def print_crack_system(directions):
    """
    Pretty printing of crack crystallographic coordinate system

    Specified by list of Miller indices for crack_direction (x),
    cleavage_plane (y) and crack_front (z), each of which should be
    a sequence of three floats
    """

    crack_direction, cleavage_plane, crack_front = directions

    crack_direction = MillerDirection(crack_direction)
    cleavage_plane = MillerPlane(cleavage_plane)
    crack_front = MillerDirection(crack_front)

    print('Crack system              %s%s' % (cleavage_plane, crack_front))
    print('Crack direction (x-axis)  %s' % crack_direction)
    print('Cleavage plane  (y-axis)  %s' % cleavage_plane)
    print('Crack front     (z-axis)  %s\n' % crack_front)



class ConstantStrainRate(object):
    """
    Constraint which increments epsilon_yy at a constant strain rate

    Rescaling is applied only to atoms where `mask` is True (default is all atoms)
    """

    def __init__(self, orig_height, delta_strain, mask=None):
        self.orig_height = orig_height
        self.delta_strain = delta_strain
        if mask is None:
            mask = Ellipsis
        self.mask = mask

    def adjust_forces(self, atoms, forces):
        pass

    def adjust_positions(self, atoms, newpos):
        current_height = newpos[:, 1].max() - newpos[:, 1].min()
        current_strain = current_height / self.orig_height - 1.0
        new_strain = current_strain + self.delta_strain
        alpha = (1.0 + new_strain) / (1.0 + current_strain)
        newpos[self.mask, 1] = newpos[self.mask, 1]*alpha

    def copy(self):
        return ConstantStrainRate(self.orig_height,
                                  self.delta_strain,
                                  self.mask)

    def apply_strain(self, atoms, rigid_constraints=False):
        """
        Applies a constant strain to the system.

        Parameters
        ----------
        atoms : ASE.atoms
            Atomic configuration.
        rigid_constraints : boolean
            Apply (or not apply) strain to every atom.
            i.e. allow constrainted atoms to move during strain application
        """

        if rigid_constraints == False:
            initial_constraints = atoms.constraints
            atoms.constraints = None

        newpos = atoms.get_positions()
        self.adjust_positions(atoms, newpos)
        atoms.set_positions(newpos)

        if rigid_constraints == False:
            atoms.constraints = initial_constraints



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

import random
import timeit
import os
import ase.io
from ..cluster_data import ClusterData


random.seed(123)


def worker_populate_cluster_data(rank, size,
                                 nProcLocal=None,
                                 atomic_cluster=None,
                                 clusterIndexes=None,
                                 nAtoms=None,
                                 qmCalculator=None,
                                 sharedList=None,
                                 debug_qm_calculator=False):
    """Function to calcuate total energy with TB

    Parameters
    ----------
    rank : int
        process number
    size : int
        total number of processes
    nProcLocal : int
        number of CPUS to be used for this calculation
    atomic_cluster : ASE.atoms
        Stucture on which ot perform the evaluation
    clusterIndexes : np.array
        list with indexes of different cluster atoms
    nAtoms : int
        number of atoms in the cluster
    qmCalculator : ASE.calculator
        calculator to be used for the evaluation
    sharedList : list
        mp shared list used ot store output data
    debug_qm_calculator : bool
        run the simulation in debug mode
    """

    # ------ MultiProcessing library pickes all objects and
    # ------ each workr thread recieves a copy

    # If a caluclator has the options, set parallel parameters
    try:
        # Create a new calculation seed
        qmCalculator.calculationSeed = str(int(random.random() * 1e7)) + str(rank)
        # Set OMP values for the potential
        qmCalculator.omp_set_threads = True
        qmCalculator.omp_num_threads = nProcLocal
    except AttributeError:
        pass

    # Create a cluster data object with relevan values
    mark = np.zeros(nAtoms, dtype=int)
    full_qm_forces = np.zeros((nAtoms, 3))

    if (debug_qm_calculator):
        ase.io.write("cluster_ext_" + str(rank) + ".xyz", atomic_cluster, format="extxyz")

    # ------ Run the calculation
    if (debug_qm_calculator):
        print("Starting evaluation of cluster %d :parallelTime" %
              (rank))
    t0 = timeit.default_timer()
    qm_forces_array = qmCalculator.get_forces(atomic_cluster)
    t1 = timeit.default_timer()
    if (debug_qm_calculator):
        print("Time taken for cluster %d: %.7e  w %d atoms :parallelTime" %
              (rank, (t1 - t0), len(atomic_cluster)))

    for i in range(atomic_cluster.info["no_quantum_atoms"]):
        orig_index = atomic_cluster.arrays["orig_index"][i]
        full_qm_forces[orig_index, :] = qm_forces_array[i, :]
        mark[orig_index] = atomic_cluster.arrays["cluster_mark"][i]

    cluster_data = ClusterData(nAtoms, mark, clusterIndexes, full_qm_forces)
    # Try to add additional details to the cluster data
    cluster_data.nClusterAtoms = len(atomic_cluster)
    qm_charges = np.zeros(nAtoms) - 10
    try:
        for i in range(atomic_cluster.info["no_quantum_atoms"]):
            orig_index = atomic_cluster.arrays["orig_index"][i]
            qm_charges[orig_index] = qmCalculator.results["charges"][i]
    except KeyError:
        pass
    cluster_data.qm_charges = qm_charges
    sharedList[rank] = cluster_data

    if (debug_qm_calculator):
        try:
            atomic_cluster.arrays["qm_charges"] = qmCalculator.results["charges"].copy()
            atomic_cluster.arrays["qm_forces"] = qmCalculator.results["forces"].copy()
        except KeyError:
            pass
        ase.io.write("cluster_ext_" + str(rank) + ".xyz", atomic_cluster, format="extxyz")



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#                  Lukas Elflein, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
from .poisson_boltzmann_distribution import ionic_strength, debye
from .poisson_nernst_planck_solver import PoissonNernstPlanckSystem
from .continuous2discrete import generate_structure as continuous2discrete
from .continuous2discrete import get_histogram



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2020) Alexander Held,
#                  Thomas Reichenbach
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""

Classes to be used with ASE in order to perform pressure relaxation and/or sliding simulations under pressure.

A usage example is found in the example directory.

Some parts are based on L. Pastewka, S. Moser, and M. Moseler, Tribol. Lett. 39, 49 (2010)
as indicated again below.

"""

from __future__ import (
    division,
    absolute_import,
    print_function,
    unicode_literals
)
import logging
import numpy as np
from ase.units import kB, fs, GPa


logger = logging.getLogger(__name__)


class AutoDamping(object):
    """Automatic damping.

    Following L. Pastewka, S. Moser, and M. Moseler,
    Tribol. Lett. 39, 49 (2010).

    Parameters
    ----------
    C11 : float
        Elastic material constant.
    p_c : float
        Empirical cut-off parameter.
    """

    def __init__(self, C11, p_c=0.01):
        self.C11 = float(C11)
        self.p_c = float(p_c)

    def get_M_gamma(self, slider, atoms):
        """Calculate mass M and dissipation constant gamma.

        Parameters
        ----------
        slider : matscipy.pressurecoupling.SlideWithNormalPressureCuboidCell
            ASE constraint used for sliding with pressure coupling.
        atoms : ase.Atoms
            Atomic configuration.

        Returns
        -------
        M: float
            Mass parameter.
        gamma : float
            Dissipation constant parameter.
        """
        A = slider.get_A(atoms)
        l = atoms.cell[slider.vdir, slider.vdir]
        t_c = l / slider.v
        omega_c = 2. * np.pi / t_c
        h1 = atoms.positions[slider.top_mask, slider.Pdir].min()
        h2 = atoms.positions[slider.bottom_mask, slider.Pdir].max()
        h = h1 - h2
        k = self.C11 * A / h
        M = k * omega_c ** -2 * np.sqrt(self.p_c ** -2 - 1.)
        # gamma = np.sqrt(2. * M * k)  # <- expression from paper, but this is wrong
        gamma = 2. * np.sqrt(M * k)
        return M, gamma


class FixedDamping(object):
    """Damping with fixed damping constant and fixed mass.

    Parameters
    ----------
    gamma : float
        Damping constant.
    M_factor : float
        Multiplicative factor to increase actual mass of upper rigid atoms.
    """

    def __init__(self, gamma, M_factor=1.0):
        self.gamma = float(gamma)
        self.M_factor = float(M_factor)

    def get_M_gamma(self, slider, atoms):
        """Calculate mass M and damping constant gamma.

        Parameters
        ----------
        slider : matscipy.pressurecoupling.SlideWithNormalPressureCuboidCell
            ASE constraint used for sliding with pressure coupling.
        atoms : ase.Atoms
            Atomic configuration.

        Returns
        -------
        M: float
            Mass parameter.
        gamma : float
            Damping parameter.
        """
        M_top = atoms.get_masses()[slider.top_mask].sum()
        return M_top * self.M_factor, self.gamma


class FixedMassCriticalDamping(object):
    """Damping with fixed mass and critical damping constant.

    Useful for fast pressure equilibration with small lid mass.

    Parameters
    ----------
    C11 : float
        Elastic material constant.
    M_factor : float
        Multiplicative factor to increase actual mass of upper rigid atoms.
    """

    def __init__(self, C11, M_factor=1.0):
        self.C11 = float(C11)
        self.M_factor = float(M_factor)

    def get_M_gamma(self, slider, atoms):
        """Calculate mass M and damping constant gamma.

        Parameters
        ----------
        slider : matscipy.pressurecoupling.SlideWithNormalPressureCuboidCell
            ASE constraint used for sliding with pressure coupling.
        atoms : ase.Atoms
            Atomic configuration.

        Returns
        -------
        M: float
            Mass parameter.
        gamma : float
            Damping parameter.
        """
        M_top = atoms.get_masses()[slider.top_mask].sum()
        M = M_top * self.M_factor
        A = slider.get_A(atoms)
        h1 = atoms.positions[slider.top_mask, slider.Pdir].min()
        h2 = atoms.positions[slider.bottom_mask, slider.Pdir].max()
        h = h1 - h2
        k = self.C11 * A / h
        gamma = 2. * np.sqrt(M * k)
        return M, gamma


class SlideWithNormalPressureCuboidCell(object):
    """ASE constraint used for sliding with pressure coupling.

    Following L. Pastewka, S. Moser, and M. Moseler,
    Tribol. Lett. 39, 49 (2010).

    Parameters
    ----------
    top_mask : boolean numpy array
        Array a with a[i] == True for each index i of the
        constraint top atoms (the atoms which slide with constant speed).
    bottom_mask : boolean numpy array
        same as top_mask but for completely fixed bottom atoms.
    Pdir : int
        Index of cell axis (0, 1, 2) along which normal pressure is applied.
    P : int
        Normal pressure in ASE units (e.g. 10.0 * ase.units.GPa).
    vdir : int
        Index of cell axis (0, 1, 2) along which to slide.
    v : float
        Constant sliding speed in ASE units (e.g. 100.0 * ase.units.m / ase.units.s).
    damping :
        Damping object (e.g. matscipy.pressurecoupling.AutoDamping instance).
    """

    def __init__(self, top_mask, bottom_mask, Pdir, P, vdir, v, damping):
        self.top_mask = top_mask
        self.Ntop = top_mask.sum()
        self.bottom_mask = bottom_mask
        self.Pdir = int(Pdir)
        self.P = float(P)
        self.vdir = int(vdir)
        self.v = float(v)
        self.damping = damping

    @property
    def Tdir(self):
        """Direction used for thermostatting.

        Thermostat direction is normal to the sliding direction and the direction of the
        applied load direction.

        Returns
        -------
        int
            Direction used for thermostatting.
        """
        all_dirs = {0, 1, 2}
        all_dirs.remove(self.Pdir)
        all_dirs.remove(self.vdir)
        return all_dirs.pop()

    @property
    def middle_mask(self):
        """Mask of free atoms.

        Returns
        -------
        numpy boolean array
            Array a with a[i] == True for each index i of the atoms
            not being part of lower or upper rigid group.
        """
        return np.logical_not(np.logical_or(self.top_mask, self.bottom_mask))

    def adjust_positions(self, atoms, positions):
        """Do not adjust positions."""
        pass

    def get_A(self, atoms):
        """Calculate cell area normal to applied load.

        Returns
        -------
        float
            Cell area normal to applied load.

        Raises
        ------
        NotImplementedError
            If atoms.get_cell() is non-orthogonal,
            SlideWithNormalPressureCuboidCell only works for orthogonal cells.
        """
        if np.abs(atoms.get_cell().sum() - atoms.get_cell().trace()) > 0:
            raise NotImplementedError("Can't do non-orthogonal cell!")
        A = 1.0
        for c in (0, 1, 2):
            if c != self.Pdir:
                A *= atoms.cell[c, c]
        return A

    def adjust_forces(self, atoms, forces):
        """Adjust forces of upper and lower rigid atoms.

        Raises
        ------
        NotImplementedError
            If atoms.get_cell() is non-orthogonal,
            SlideWithNormalPressureCuboidCell only works for orthogonal cells.
        """
        if np.abs(atoms.get_cell().sum() - atoms.get_cell().trace()) > 0:
            raise NotImplementedError("Can't do non-orthogonal cell!")
        A = self.get_A(atoms)
        M, gamma = self.damping.get_M_gamma(self, atoms)
        Ftop = forces[self.top_mask, self.Pdir].sum()
        vtop = atoms.get_velocities()[self.top_mask, self.Pdir].sum() / self.Ntop
        F = Ftop - self.P * A - gamma * vtop
        a = F / M
        forces[self.bottom_mask, :] = .0
        forces[self.top_mask, :] = .0
        forces[self.top_mask, self.Pdir] = atoms.get_masses()[self.top_mask] * a

    def adjust_momenta(self, atoms, momenta):
        """Adjust momenta of upper and lower rigid atoms.

        Raises
        ------
        NotImplementedError
            If atoms.get_cell() is non-orthogonal,
            SlideWithNormalPressureCuboidCell only works for orthogonal cells.
        """
        if np.abs(atoms.get_cell().sum() - atoms.get_cell().trace()) > 0:
            raise NotImplementedError("Can't do non-orthogonal cell!")
        top_masses = atoms.get_masses()[self.top_mask]
        vtop = (momenta[self.top_mask, self.Pdir] / top_masses).sum() / self.Ntop
        momenta[self.bottom_mask, :] = 0.0
        momenta[self.top_mask, :] = 0.0
        momenta[self.top_mask, self.vdir] = self.v * top_masses
        momenta[self.top_mask, self.Pdir] = vtop * top_masses

    def adjust_potential_energy(self, atoms):
        """Do not adjust energy."""
        return 0.0


class SlideLogger(object):
    """Logger to be attached to an ASE integrator.

    For new files (not restart jobs), the write_header method should
    be called once in order to write the header to the file. Also note
    that ASE does not write the time step 0 automatically.
    You can do so by calling the SlideLogger instance once
    before starting the integration as in example 1 below.

    Parameters
    ----------
    handle : filehandle
       Filehandle e.g. pointing to a file opened in w or a mode.
    atoms : ase.Atoms
        Atomic configuration.
    slider : slider object
        Instance of SlideWithNormalPressureCuboidCell.
    integrator : ASE integrator object,
        Instance of ASE integrator e.g. ase.md.langevin.Langevin.
    step_offset : int
        Last step already written to log file, useful for restarts.

    Examples
    --------
    1. For new runs:
        log_handle = open(logfn, 'w', 1)  # line buffered
        logger = SlideLogger(log_handle, ...)
        logger.write_header()
        logger()
        integrator.attach(logger)
        integrator.run(steps_integrate)
        log_handle.close()

    2. For restarts:
        with open(logfn, 'r') as log_handle:
            step_offset = SlideLog(log_handle).step[-1]
        log_handle = open(logfn, 'a', 1)  # line buffered append
        logger = SlideLogger(log_handle, ..., step_offset=step_offset)
        integrator.attach(logger)
        integrator.run(steps_integrate)
        log_handle.close()
    """

    def __init__(self, handle, atoms, slider, integrator, step_offset=0):
        self.handle = handle
        self.atoms = atoms
        self.slider = slider
        self.integrator = integrator
        self.step_offset = step_offset

    def write_header(self):
        """Write header of log-file."""
        self.handle.write('# step | time / fs | T_thermostat / K | P_top / GPa | P_bottom / GPa | h / Ang | v / Ang * fs | a / Ang * fs ** 2 | tau_top / GPa | tau_bottom / GPa\n')

    def __call__(self):
        """Write current status (time, T, P, ...) to log-file."""
        slider = self.slider
        atoms = self.atoms
        integrator = self.integrator
        p = atoms.get_momenta()[slider.middle_mask, slider.Tdir]
        v = atoms.get_velocities()[slider.middle_mask, slider.Tdir]
        dof = len(p)
        Ekin = 0.5 * (p * v).sum()
        T = 2. * Ekin / (dof * kB)
        step = integrator.nsteps + self.step_offset
        t = step * integrator.dt / fs
        A = atoms.cell[slider.vdir, slider.vdir] * atoms.cell[slider.Tdir, slider.Tdir]
        F = atoms.get_forces(apply_constraint=False)
        F_top = F[slider.top_mask, slider.Pdir].sum()
        F_bottom = F[slider.bottom_mask, slider.Pdir].sum()
        P_top = F_top / A / GPa
        P_bottom = F_bottom / A / GPa
        h1 = atoms.positions[slider.top_mask, slider.Pdir].min()
        h2 = atoms.positions[slider.bottom_mask, slider.Pdir].max()
        h = h1 - h2
        v = atoms.get_velocities()[slider.top_mask, slider.Pdir][0] * fs
        a = atoms.get_forces(md=True)[slider.top_mask, slider.Pdir][0] / atoms.get_masses()[slider.top_mask][0] * fs ** 2
        F_v_top = F[slider.top_mask, slider.vdir].sum()
        tau_top = F_v_top / A / GPa
        F_v_bottom = F[slider.bottom_mask, slider.vdir].sum()
        tau_bottom = F_v_bottom / A / GPa
        self.handle.write('%d   %r   %r   %r   %r   %r   %r   %r   %r   %r\n' % (step, float(t), float(T), float(P_top), float(P_bottom), float(h), float(v), float(a), float(tau_top), float(tau_bottom)))


class SlideLog(object):
    """Reader for logs written with SlideLogger instance.

    Parameters
    ----------
    handle matscipy.pressurecoupling.SlideLogger instance
        Handle or filename pointing to log file.

    Attributes
    ----------
    step : ndarray
        Step indices 0, 1, 2, ....
    time : ndarray
        Simulation time in fs at step.
    T_thermostat : ndarray
        Instantantaneous temperature in K from thermostat-region
        only from degrees of freedom along thermalized direction.
    P_top : ndarray
        Normal pressure on lid in GPa.
    P_bottom : ndarray
        Normal pressure on base in GPa.
    h : ndarray
        Separation of lid and base in Ang.
    v : ndarray
        Normal speed of lid in Ang / fs.
    a : ndarray
        Normal acceleration of lid in Ang / fs ** 2.
    tau_top : ndarray
        Shear stress on lid in GPa.
    tau_bottom : ndarray
        Shear stress on base in GPa.
    rows : ndarray
        All data in a 2d array with axis 0 step and axis 1
        the values ordered as above.
    """

    def __init__(self, handle):
        self.rows = np.loadtxt(handle)
        self.step, self.time, self.T_thermostat, self.P_top, self.P_bottom, self.h, self.v, self.a, self.tau_top, self.tau_bottom = self.rows.T
        self.step = self.step.astype(int)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) Till Junge, Karlsruhe Institute of Technology
#                  James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import multiprocessing
import multiprocessing.managers

try:
    import argparse
except ImportError:
    from matscipy.logger import screen
    screen.pr('argparse module not availability, some functionality disabled')

import abc
import datetime
import sys


class BaseResultManager(object):
    """
    Baseclass for job distribution servers. User needs to implement the method process
    """
    __metaclass__ = abc.ABCMeta
    def __init__(self, port, key):
        """

        Keyword Arguments:
        port    -- listening port
        key     -- auth_key
        verbose -- (default False) if set, outputs debugging messages
        """
        self.port = port
        self.key = key
        self.job_queue = None
        self.result_queue = None
        self.todo_counter = None
        self.work_done_flag = None
        self.manager = None
        self.done = False

        self.create_manager()

    def create_manager(self):
        """
        creates a multiprocessing.SyncManager
        """
        self.job_queue = multiprocessing.JoinableQueue()
        self.result_queue = multiprocessing.JoinableQueue()
        # the -1 is for 'uninitialized
        self.todo_counter = multiprocessing.Manager().Value('i', -1)
        self.work_done_flag = multiprocessing.Manager().Event()
        self.work_done_flag.clear()
        # This is based on the examples in the official docs of multiprocessing.
        # get_{job|result}_q return synchronized proxies for the actual Queue
        # objects.
        class JobQueueManager(multiprocessing.managers.SyncManager):
            pass

        JobQueueManager.register('get_job_queue',
                                 callable=lambda: self.job_queue)
        JobQueueManager.register('get_result_queue',
                                 callable=lambda: self.result_queue)
        JobQueueManager.register('get_todo_counter',
                                 callable=lambda: self.todo_counter,
                                 proxytype= multiprocessing.managers.ValueProxy)
        JobQueueManager.register('get_work_done_event',
                                 callable=lambda: self.work_done_flag,
                                 proxytype= multiprocessing.managers.EventProxy)
        self.manager = JobQueueManager(address=('', self.port), authkey=self.key)
        self.manager.start()

    def set_todo_counter(self, counter):
        self.todo_counter.set(counter)
        self.done = (counter == 0)

    def get_todo_counter(self):
        return self.todo_counter.get()

    def decrement_todo_counter(self):
        new_counter = self.todo_counter.get() - 1
        self.done = (new_counter == 0)
        self.todo_counter.set(self.todo_counter.get() - 1)

    @classmethod
    def get_arg_parser(cls, parser=None):
        """
        create or extend a argparser to read command line arguments required by
        the server.
        Keyword Arguments:
        parser -- optional: if provided, parser is extended to include port and
                  authentication key
        """
        if parser is None:
            parser = argparse.ArgumentParser()

        parser.add_argument('--port', type=int, default=9995,
                            help='server listening port')
        parser.add_argument('--auth-token', type=str, default='auth_token',
                            help=('shared information used to authenticate the '
                                  'client to the server'))
        return parser

    def run(self):
        """
        this is the actual serving method. it fills the jobqueue and processes
        incoming results
        """
        print("Start serving jobs and processing results")
        while not self.done:
            self.schedule_available_jobs()
            self.receive_results()
        print()
        print("Signalling end of work to worker processes")
        self.work_done_flag.set()
        print("Waiting for stragglers to hand in results")
        self.result_queue.join()
        print("Wrapping this up")
        self.manager.shutdown()

    @abc.abstractmethod
    def schedule_available_jobs(self):
        """
        to be implemented by inheriting classes. should push available jobs
        into the job queue
        """
        raise NotImplementedError()

    def receive_results(self):
        """
        proposed standard result receiver, can be overloaded by inheriting
        classes
        """
        try:
            result = self.result_queue.get()
            if result:
                value, job_id = result
                self.process(value, job_id)
        finally:
            self.result_queue.task_done()

    @abc.abstractmethod
    def process(self, value, job_id):
        """
        to be implemented by inheriting classes. should push available jobs
        into the job queue
        """
        raise NotImplementedError()


class BaseWorker(multiprocessing.Process):
    """
    Baseclass for distributed calculation worker threads
    """
    __metaclass__ = abc.ABCMeta
    def __init__(self, server_address, port, key, verbose=False, walltime=None):
        """

        Keyword Arguments:
        server_address -- ip or fully qualified hostname
        port           -- listening port
        key            -- auth_key
        verbose        -- (default False) if set, outputs debugging messages
        walltime       -- (default None) if set, worker commits suicide after
                          walltime hours
        """
        super(BaseWorker, self).__init__()
        self.server_address = server_address
        self.port = port
        self.key = key

        self.job_queue = None
        self.result_queue = None
        self.todo_counter = None
        self.work_done_flag = None
        self.manager = None

        self.create_manager()
        self.verbose = verbose
        self.commit_suicide = walltime is not None
        self.time_of_death = None
        if self.commit_suicide:
            self.time_of_death = (datetime.datetime.now() +
                                  datetime.timedelta(hours = walltime))


    def create_manager(self):
        """
        creates a multiprocessing.SyncManager
        """
        self.job_queue = multiprocessing.JoinableQueue()
        self.result_queue = multiprocessing.JoinableQueue()
        # the -1 is for 'uninitialized
        self.todo_counter = multiprocessing.Manager().Value('i', -1)
        self.work_done_flag = multiprocessing.Manager().Event()
        self.work_done_flag.clear()

        # This is based on the examples in the official docs of multiprocessing.
        # get_{job|result}_q return synchronized proxies for the actual Queue
        # objects.
        class ServerQueueManager(multiprocessing.managers.SyncManager):
            pass

        ServerQueueManager.register('get_job_queue')
        ServerQueueManager.register('get_result_queue')
        ServerQueueManager.register('get_todo_counter')
        ServerQueueManager.register('get_work_done_event')

        self.manager = ServerQueueManager(
            address=(self.server_address, self.port),
             authkey=self.key)
        self.manager.connect()

        self.job_queue = self.manager.get_job_queue()
        self.result_queue = self.manager.get_result_queue()
        self.todo_counter = self.manager.get_todo_counter()
        self.work_done_flag = self.manager.get_work_done_event()
        return self.manager

    @classmethod
    def get_arg_parser(cls, parser=None):
        """
        create or extend a argparser to read command line arguments required by
        the cliend.
        Keyword Arguments:
        parser -- optional: if provided, parser is extended to include port and
                  authentication key
        """
        if parser is None:
            parser = argparse.ArgumentParser()

        parser.add_argument('--server_address', metavar='INET_ADDR', type=str,
                            default='',
                            help=('job server ip address or fully qualified '
                                  'hostname'))
        parser.add_argument('--port', type=int, default=9995,
                            help='server listening port')
        parser.add_argument('--auth-token', type=str, default='auth_token',
                            help=('shared information used to authenticate the '
                                  'client to the server'))
        return parser


    def run(self):
        """
        standard method that any multiprocessing.Process must implement
        """
        if self.verbose:
            print("Starting to run")

        if self.commit_suicide:
            def gotta_commit_suicide():
                do_I = datetime.datetime.now() > self.time_of_death
                if do_I:
                    print("Reached walltime, stopping accepting new jobs (zombie)")
                return do_I
        else:
            def gotta_commit_suicide():
                return False

        while not self.work_done_flag.is_set() and not gotta_commit_suicide():
            try:
                if self.verbose:
                    print("trying to get a job")
                job_description, job_id = self.job_queue.get()
                if self.verbose:
                    print("got job {}".format(job_id))
                try:
                    self.process(job_description, job_id)
                except Exception as err:
                    print("ERROR:::: {}".format(err))
                    raise
            finally:
                try:
                    self.job_queue.task_done()
                except EOFError:
                    pass

    @abc.abstractmethod
    def process(self, job_description, job_id):
        raise NotImplementedError()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2016) Andreas Klemenz, Fraunhofer IWM
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import time
import numpy as np

import ase
import ase.data
import ase.io
import ase.io.lammpsrun

import matscipy.neighbours
import sys


def twochar(name):
    if len(name) > 1:
        return name[:2]
    else:
        return name + ' '


class BondData:
    def __init__(self, name_value_hash=None):
        if name_value_hash:
            self.nvh = name_value_hash
            self.set_names(name_value_hash.keys())

    def set_names(self, names):
        if not hasattr(self, 'names'):
            self.names = set()
        for name in names:
            aname, bname = name.split('-')
            name1 = twochar(aname) + '-' + twochar(bname)
            name2 = twochar(bname) + '-' + twochar(aname)
            if not name1 in self.names and not name2 in self.names:
                self.names.add(name1)

    def get_name(self, aname, bname):
        name1 = twochar(aname) + '-' + twochar(bname)
        name2 = twochar(bname) + '-' + twochar(aname)
        if name1 in self.names:
            return name1
        elif name2 in self.names:
            return name2
        else:
            return None
    
    def name_value(self, aname, bname):
        name1 = twochar(aname) + '-' + twochar(bname)
        name2 = twochar(bname) + '-' + twochar(aname)
        if name1 in self.nvh:
            return name1, self.nvh[name1]
        if name2 in self.nvh:
            return name2, self.nvh[name2]
        return None, None

    def get_value(self, aname, bname):
        return self.name_value(aname, bname)[1]
        

class CutoffList(BondData):
    def max(self):
        return max(self.nvh.values())


class AnglesData:
    def __init__(self, name_value_hash=None):
        if name_value_hash:
            self.nvh = name_value_hash
            self.set_names(name_value_hash.keys())

    def set_names(self, names):
        if not hasattr(self, 'names'):
            self.names = set()
        for name in names:
            aname, bname, cname = name.split('-')
            name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname)
            name2 = twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
            if not name1 in self.names and not name2 in self.names:
                self.names.add(name1)

    def add_name(self, aname, bname, cname):
        if not hasattr(self, 'names'):
            self.names = set()
        name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname)
        name2 = twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
        if not name1 in self.names and not name2 in self.names:
            self.names.add(name1)

    def get_name(self, aname, bname, cname):
        if not hasattr(self, 'names'):
            return None
        name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname)
        name2 = twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
        if name1 in self.names:
            return name1
        elif name2 in self.names:
            return name2
        else:
            return None
    
    def name_value(self, aname, bname, cname):
        for name in [
            (twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname)),
            (twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname))]:
            if name in self.nvh:
                return name, self.nvh[name]
        return None, None
    

class DihedralsData:
    def __init__(self, name_value_hash=None):
        if name_value_hash:
            self.nvh = name_value_hash
            self.set_names(name_value_hash.keys())

    def set_names(self, names):
        if not hasattr(self, 'names'):
            self.names = set()
        for name in names:
            aname, bname, cname, dname = name.split('-')
            name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname) + '-' + twochar(dname)
            name2 = twochar(dname) + '-' + twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
            if not name1 in self.names and not name2 in self.names:
                self.names.add(name1)

    def add_name(self, aname, bname, cname, dname):
        if not hasattr(self, 'names'):
            self.names = set()
        name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname) + '-' + twochar(dname)
        name2 = twochar(dname) + '-' + twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
        if not name1 in self.names and not name2 in self.names:
            self.names.add(name1)

    def get_name(self, aname, bname, cname, dname):
        if not hasattr(self, 'names'):
            return None
        name1 = twochar(aname) + '-' + twochar(bname) + '-' + twochar(cname) + '-' + twochar(dname)
        name2 = twochar(dname) + '-' + twochar(cname) + '-' + twochar(bname) + '-' + twochar(aname)
        if name1 in self.names:
            return name1
        elif name2 in self.names:
            return name2
        else:
            return None
    
    def name_value(self, aname, bname, cname, dname):
        for name in [
            (twochar(aname) + '-' + twochar(bname) + '-' +
             twochar(cname) + '-' + twochar(dname)),
            (twochar(dname) + '-' + twochar(cname) + '-' + 
             twochar(bname) + '-' + twochar(aname))]:
            if name in self.nvh:
                return name, self.nvh[name]
        return None, None



class OPLSStructure(ase.Atoms):
    default_map = {
        'BR': 'Br',
        'Be': 'Be',
        'C0': 'Ca',
        'Li': 'Li',
        'Mg': 'Mg',
        'Al': 'Al',
        'Ar': 'Ar',
        }

    def __init__(self, *args, **kwargs):
        ase.Atoms.__init__(self, *args, **kwargs)
        if len(self) == 0:
            self.types = None
        else:
            types = np.array(self.get_chemical_symbols())
            self.types = np.unique(types)

            tags = np.zeros(len(self), dtype=int)
            for it, type in enumerate(self.types):
                tags[types == type] = it
            self.set_tags(tags)

        self._combinations = {}
        self._combinations[0] = []
        self._combinations[1] = []
        self._combinations[2] = [(0, 1)]
        self._combinations[3] = [(0, 1), (0, 2), (1, 2)]
        self._combinations[4] = [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3)]
        self._combinations[5] = [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3), (0, 4), (1, 4), (2, 4), (3, 4)]
        self._combinations[6] = [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3), (0, 4), (1, 4), (2, 4), (3, 4), (0, 5), (1, 5), (2, 5), (3, 5), (4, 5)]
        self._combinations[7] = [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3), (0, 4), (1, 4), (2, 4), (3, 4), (0, 5), (1, 5), (2, 5), (3, 5), (4, 5), (0, 6), (1, 6), (2, 6), (3, 6), (4, 6), (5, 6)]
        self._combinations[8] = [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3), (0, 4), (1, 4), (2, 4), (3, 4), (0, 5), (1, 5), (2, 5), (3, 5), (4, 5), (0, 6), (1, 6), (2, 6), (3, 6), (4, 6), (5, 6), (0, 7), (1, 7), (2, 7), (3, 7), (4, 7), (5, 7), (6, 7)]

    def _get_combinations(self, n):
        r = range(n)
        i = np.tile(r,n)
        j = np.repeat(r,n)
        return zip(i[i < j], j[i < j])

    def append(self, atom):
        """Append atom to end."""
        self.extend(ase.Atoms([atom]))

    def set_types(self, types):
        types = np.array(types)
        self.types = np.unique(types)

        tags = np.zeros(len(self), dtype=int)
        for it, type in enumerate(self.types):
            tags[types == type] = it
        self.set_tags(tags)

    def get_types(self):
        return self.types

    def set_cutoffs(self, cutoffs):
        self.cutoffs = cutoffs

    def get_neighbors(self):
        atoms = ase.Atoms(self)
        types = np.array(self.get_types())
        tags = atoms.get_tags()

        cut = self.cutoffs.max()
        ni, nj, dr = matscipy.neighbours.neighbour_list('ijd', atoms, cut)

        tags2cutoffs = np.full([len(types), len(types)], -1.)
        for i, itype in enumerate(types):
            for j, jtype in enumerate(types):
                tags2cutoffs[i,j] = self.cutoffs.get_value(itype, jtype)

        cutoff_undef = np.where(tags2cutoffs < 0.)
        if np.shape(cutoff_undef)[1] > 0:
            for i in range(np.shape(cutoff_undef)[1]):
                iname = types[cutoff_undef[0][i]]
                jname = types[cutoff_undef[1][i]]
                print('ERROR: Cutoff %s-%s not found' % (iname, jname))
                sys.exit()
        cut = tags2cutoffs[tags[ni], tags[nj]]

        self.ibond = ni[dr <= cut]
        self.jbond = nj[dr <= cut]


    def set_atom_data(self, atom_data):
        self.atom_data = atom_data


    def get_charges(self):
        types = self.types[self.get_tags()]
        self.charges = np.zeros(len(self))

        for i, itype in enumerate(self.types):
            self.charges[self.get_tags() == i] = self.atom_data[itype][2]

        return self.charges


    def get_bonds(self, bonds=None):
        types = np.array(self.get_types())
        tags = self.get_tags()

        if bonds:
            self.bonds = bonds

        if not hasattr(self, 'ibond'):
            self.get_neighbors()
        ibond = self.ibond
        jbond = self.jbond

        # remove duplicates from neighbor list
        mask = jbond <= ibond
        ibond = ibond[mask]
        jbond = jbond[mask]


        tags2bond_names = np.full([len(types), len(types)], '', dtype=object)
        for i, itype in enumerate(types):
            for j, jtype in enumerate(types):
                name = self.cutoffs.get_name(itype, jtype)
                if not name:
                    print('ERROR: Cutoff %s-%s not found' % (itype, jtype))
                    sys.exit()
                tags2bond_names[i,j] = name

        names = tags2bond_names[tags[ibond], tags[jbond]]

        self.bond_types = np.unique(names)

        self.bond_list = np.empty([0,3], dtype=int)
        for n,bond_type in enumerate(self.bond_types):
            mask = names == bond_type
            bond_list_n = np.empty([len(np.where(mask)[0]),3], dtype=int)
            bond_list_n[:,0] = n
            bond_list_n[:,1] = ibond[np.where(mask)]
            bond_list_n[:,2] = jbond[np.where(mask)]
            self.bond_list = np.append(self.bond_list, bond_list_n, axis=0)

        if hasattr(self, 'bonds'):
            potential_unknown = False
            for nb, bond_type in enumerate(self.bond_types):
                itype, jtype = bond_type.split('-')
                if not self.bonds.get_value(itype, jtype):
                    print('ERROR: Pair potential %s-%s not found' % (itype, jtype))
                    print('List of affected bonds:')
                    mask = self.bond_list.T[0] == nb
                    print(self.bond_list[mask].T[1:].T)
                    potential_unknown = True
            if potential_unknown:
                sys.exit()

        return self.bond_types, self.bond_list


    def get_angles(self, angles=None):
        types = np.array(self.get_types())
        tags = self.get_tags()
         
        self.ang_list = []
        self.ang_types = []

        if not hasattr(self, 'ibond'):
            self.get_neighbors()
        ibond = self.ibond
        jbond = self.jbond

        if angles:
            self.angles = angles

        if not hasattr(self, 'angles'):
            self.angles = AnglesData()
            for itype in types:
                for jtype in types:
                    for ktype in types:
                        self.angles.add_name(itype, jtype, ktype)



        angles_undef = AnglesData()
        angles_undef_lists = {}

        for i in range(len(self)):
            iname = types[tags[i]]

            ineigh = jbond[ibond == i]
            n_ineigh = np.shape(ineigh)[0]

            if n_ineigh not in self._combinations.keys():
                self._combinations[n_ineigh] = self._get_combinations(n_ineigh)

            for nj,nk in self._combinations[len(ineigh)]:
                j = ineigh[nj]
                k = ineigh[nk]

                jname = types[tags[j]]
                kname = types[tags[k]]
                name = self.angles.get_name(jname, iname, kname)

                if hasattr(self, 'angles') and not name:
                    # Angle found without matching parameter definition
                    # Add to list anyway to get meaningful error messages
                    if not angles_undef.get_name(jname, iname, kname):
                        angles_undef.add_name(jname, iname, kname)
                        angles_undef_lists[angles_undef.get_name(jname, iname, kname)] = [[j, i, k]]
                    else:
                        angles_undef_lists[angles_undef.get_name(jname, iname, kname)].append([j, i, k])
                    continue

                if name not in self.ang_types:
                    self.ang_types.append(name)
                self.ang_list.append([self.ang_types.index(name), j, i, k])


        if len(angles_undef_lists) > 0:
            for name in angles_undef_lists:
                print('ERROR: Angular potential %s not found' % (name))
                print('List of affected angles:')
                for angle in angles_undef_lists[name]:
                    print(angle)
            sys.exit()

        return self.ang_types, self.ang_list



    def get_dihedrals(self, dihedrals=None, full_output=False):
        types = self.get_types()
        tags = self.get_tags()

        if dihedrals:
            self.dihedrals = dihedrals

        if not hasattr(self, 'dihedrals'):
            self.dihedrals = DihedralsData()
            for itype in types:
                for jtype in types:
                    for ktype in types:
                        for ltype in types:
                            self.dihedrals.add_name(itype, jtype, ktype, ltype)


        self.dih_list = []
        self.dih_types = []

        if not hasattr(self, 'ibond'):
            self.get_neighbors()
        ibond = self.ibond
        jbond = self.jbond


        dihedrals_undef_lists = {}

        for j,k in zip(ibond, jbond):
            if j < k:
                jname = types[tags[j]]
                kname = types[tags[k]]

                i_dihed = jbond[ibond == j]
                l_dihed = jbond[ibond == k]
                i_dihed = i_dihed[i_dihed != k]
                l_dihed = l_dihed[l_dihed != j]

                for i in i_dihed:
                    iname = types[tags[i]]
                    for l in l_dihed:
                        lname = types[tags[l]]

                        name = self.dihedrals.get_name(iname, jname, kname, lname)

                        if hasattr(self, 'dihedrals') and not name:
                            # Dihedral found without matching parameter definition
                            # Add to list anyway to get meaningful error messages
                            name = iname + '-' + jname + '-' + kname + '-' + lname
                            if name not in dihedrals_undef_lists:
                                dihedrals_undef_lists[name] = [[i, j, k, l]]
                            else:
                                dihedrals_undef_lists[name].append([i, j, k, l])
                            continue

                        if name not in self.dih_types:
                            self.dih_types.append(name)
                        self.dih_list.append([self.dih_types.index(name), i, j, k, l])


        if len(dihedrals_undef_lists) > 0:
            # "dihedrals_undef_lists" might contain duplicates, i.e. A-B-C-D and D-C-B-A.
            # This could be avoided by using a "DihedralsData" object to store dihedral
            # names, similar to the way the "AnglesData" object is used in the
            # "get_angles()" method. For performance reasons, this is not done here.
            for name in dihedrals_undef_lists:
                print('WARNING: Dihedral potential %s not found' % (name))
                if not full_output:
                    print('Example for affected atoms: %s' % (str(dihedrals_undef_lists[name][0])))
                else:
                    print('Full list of affected atoms:')
                    for dihed in dihedrals_undef_lists[name]:
                        print(dihed)

        return self.dih_types, self.dih_list





from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


class ClusterData(object):
    """Class for storing cluster data

    Attributes
    ----------
    forces : np.array
        Atomic forces
    mark : list
        Marks assigning atoms as:
            1: core QM region
            2: buffer region
            3: terminal atoms (final atom included in the buffer region)
            4: additional terminal atoms
            5: Hydrogens used ot terminate cut-off bonds
    qm_list : list
        list of inner QM atoms
    """

    def __init__(self, nAtoms, mark=None, qm_list=None, forces=None):
        if len(mark) != nAtoms:
            raise ValueError(
                "mark length not compatible with atoms length in this ClusterData object")
        if np.shape(forces) != (nAtoms, 3):
            raise ValueError(
                "forces shape not compatible with atoms length in this ClusterData object")

        self.forces = forces
        self.mark = mark
        self.qm_list = qm_list
        self.nClusterAtoms = None

    def __str__(self):
        return str(self.mark)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np

import _matscipy

###

def angle_distribution(i, j, dr, nbins, *args):
    """
    Compute a bond angle distribution from a neighbour list.

    Parameters
    ----------
    i, j, dr : array_like
        Neighbour list, including list of distance vectors.
    nbins : int
        Number of bins for bond angle histogram.
    cutoff : float, optional
        Bond length cutoff, i.e. consider only bonds shorter than this length.
    """
    return _matscipy.angle_distribution(np.asarray(i), np.asarray(j),
                                        np.asarray(dr), nbins, *args)
